package hanoi;

/**
 *
 * @author Bright
 * Version 2
 * Update date: march 20, 2017
 */

// Rules :
//     河內塔問題目的乃在三根柱子中,將n個盤子從
//     A 柱子搬到 C 柱中,每次只移動一盤子,而且必須遵守
//     每個盤子都比其上面的盤子還要大的原則。
//
// Ans :
//     河內塔問題的想法必須針對最底端的盤子。
//     我們必須先把A柱子頂端n-1個盤子想辦法(借助C柱)移至B柱子
//     然後才能將想最底端的盤子移至C柱。
//     此時C有最大的盤子,B總共n-1個盤子,A柱則無。
//     只要再借助A柱子，將B柱n-1個盤子移往C柱即可 :
//
//     HanoiTower(n-1,A,C,B) ;
//     將A頂端n-1個盤子借助C移至B
//     HanoiTower(n-1,B,A,C) ;
//     將B上的n-1個盤子借助A移至C

import java.io.*;
import java.util.Scanner;
import java.util.InputMismatchException;

class Hanoi
{
  	// 遞迴函數呼叫求河內塔之解
	public static void HanoiTower(int n, char from, char aux, char to)
	{
		if (n == 1)
			System.out.printf("Move disk %d from %c --> %c\n", n, from, to);
		else {
			// 將A上n-1個盤子借助C移至B
			HanoiTower(n-1, from, to, aux) ;
			System.out.printf("Move disk %d from %c --> %c\n", n, from, to);	 
			// 將B上n-1個盤子借助A移至C
			HanoiTower(n-1, aux, from, to) ;
		}
	}
   
	public static void main (String args[]) //主函數		
	{ 
		Scanner keyboard = new Scanner(System.in);
		Hanoi obj = new Hanoi();
		int n = 0 ;
		String ch = "";
		char A='A', B='B', C='C';

		System.out.print("-----Hanoi Tower Implementaion----\n");
		// 輸入共有幾個盤子在A柱子中
		do {
		    System.out.printf("How many disks in A? ");
 	 	    try {
			   n = keyboard.nextInt();
		    } catch(InputMismatchException e) {
			   keyboard.nextLine();
			   System.out.printf("Not a correctly number.\n");
			   System.out.printf("Try again\n\n");
		    }
		    if (n == 0)
			   System.out.print("no disk to move\n");
		    else
			   obj.HanoiTower(n,A,B,C) ;
		    System.out.printf("\nContinue (y/n)? ");
		    ch = keyboard.next();
		} while (ch.equals("y"));

	}   
}
