package binarysearchtree;

/**
 *
 * @author Bright
 */

import java.io.*;
import java.util.Scanner;
import java.util.InputMismatchException;

class Student {
	String name;    // 學生姓名
	int score;      // 學生成績
	Student llink;  // 左子鏈結
	Student rlink;  // 右子鏈結
};

public class BinarySearchTree {

	Student root, ptr;
	Scanner keyboard = new Scanner(System.in);

	// 建構函數
	BinarySearchTree()
	{
		root = null;
	}

	// 新增函數; 新增一筆新的資料
	public void insert_f()
	{
		String name = "";
		int score = 0;
		System.out.printf("\n=====INSERT DATA=====\n");
		System.out.print("Enter student name: ");
		name = keyboard.next();
	
		System.out.print("Enter student score: ");
		score = keyboard.nextInt();
		
 		access(name, score);
	}

	// 刪除函數; 將資料從二元搜尋樹中刪除
	public void delete_f()
	{
		String name = "";
		if (root == null) {
			System.out.println("No student record!");
			return;
		}
		System.out.printf("\n=====DELETE DATA=====\n");
		System.out.print("Enter student name: ");
		name = keyboard.next();

		removing(name);
	}

	// 修改函數; 修改學生成績
	public void modify_f()
	{
		Student node;
		String name = "";
		if (root == null) { // 判斷根節點是否為null
			System.out.println("No student record!");
			return;
		}
		System.out.printf("\n=====MODIFY DATA=====\n");
		System.out.print("Enter student name: ");
		name = keyboard.next();

		if ((node = search(name)) == null)
			System.out.println("Student " + name + " not found!");
		else {
			// 列出原資料狀況
			System.out.println("Original student name: " + node.name);
			System.out.println("Original student score: " + node.score);
			System.out.print("Enter new score: ");
			node.score = keyboard.nextInt();
 
			System.out.println("Student " + name + " has been modified");
		}
	}

	// 輸出函數; 依照人名由小至大輸出至螢幕
	public void show_f()
	{
		if (root == null) { // 判斷根節點是否為null
			System.out.println("No student record!");
			return;
		}
		System.out.printf("\n=====SHOW DATA=====\n");
		inorder(root);  // 以中序法輸出資料
	}

	// 處理二元搜尋樹，將新增資料加入至二元搜尋樹中
	public void access(String name, int score)
	{
		Student node, prev = null;
		if (search(name) != null) { // 資料已存在則顯示錯誤
			System.out.println("Student " + name + " has existed!");
			return;
		}
		ptr = new Student();
		ptr.name = name;
		ptr.score = score;
		ptr.llink = ptr.rlink = null;
		if (root == null)  // 當根節點為null的狀況
			root = ptr;
		else { // 當根節點不為null的狀況
			node = root;
			while (node != null) {     // 搜尋資料插入點
				prev = node;
				if (ptr.name.compareTo(node.name) < 0)
					node = node.llink;
				else
					node = node.rlink;
			}
			if (ptr.name.compareTo(prev.name) < 0)
				prev.llink = ptr;
			else
				prev.rlink = ptr;
		}
	}

	// 將資料從二元搜尋樹中移除
	public void removing(String name)
	{
		Student del_node;
		if ((del_node = search(name)) == null) { // 找不到資料則顯示錯誤
			System.out.println("Student " + name + " not found!");
			return;
		}
		 // 節點不為樹葉節點的狀況
		if (del_node.llink != null || del_node.rlink != null)
			del_node = replace(del_node);
		else  // 節點為樹葉節點的狀況
			if (del_node == root)
				root = null;
			else
				connect(del_node, 'n');
		del_node = null;  // 釋放記憶體
		System.out.println("Student " + name + " has been deleted!");
	}

	// 尋找刪除非樹葉節點的替代節點
	public Student replace(Student node)
	{
		Student re_node;
		// 當右子樹找不到替代節點，會搜尋左子樹是否存在替代節點
		if ((re_node = search_re_r(node.rlink)) == null)
			re_node = search_re_l(node.llink);
		if (re_node.rlink != null)  // 當替代節點有右子樹存在的狀況
			connect(re_node, 'r');
		else
			if (re_node.llink != null)  // 當替代節點有左子樹存在的狀況
				connect(re_node, 'l');
			else  // 當替代節點為樹葉節點的狀況
				connect(re_node, 'n');
		node.name = re_node.name;
		node.score = re_node.score;
		return re_node;
	}

	// 調整二元搜尋樹的鏈結，link為r表示處理右鏈結，為l表處理左鏈結，
	// 為m則將鏈結指向null
	public void connect(Student node, char link)
	{
		Student parent;
		parent = search_p(node);  // 搜尋父節點
		if (node.name.compareTo(parent.name) < 0)  // 節點為父節點左子樹的狀況
			if (link == 'r')  /* link 為 r */
				parent.llink = node.rlink;
			else if (link == 'l')  /* link 為l */
				parent.llink = node.llink;
			else  /* link 為 n */
				parent.llink = null;
    
         /* node 節點為父節點右子樹的狀況 */
         else if (link == 'r')  /* link 為 r */
                    parent.rlink = node.rlink;
                else if (link == 'l')  /* link 為 l */
                     parent.rlink = node.llink;
                else  /* link 為 n */
                      parent.rlink = null;
	}

	// 以中序法輸出資料，採遞迴方式
	public void inorder(Student node)
	{
		if (node != null) {
			inorder(node.llink);
			System.out.printf("%-15s %-3d\n", node.name, node.score);
			inorder(node.rlink);
		}
	}

	// 搜尋target所在節點
	public Student search(String target)
	{
		Student node;
		node = root;
		while (node != null) {
			if (target.compareTo(node.name) == 0)
				return node;
			else
				if (target.compareTo(node.name) < 0)  // target小於目前節點，往左搜尋
					node = node.llink;
				else  // target大於目前節點，往右搜尋
					node = node.rlink;
		}
		return node;
	}

	// 搜尋右子樹替代節點
	Student search_re_r(Student node)
	{
		Student re_node;
		re_node = node;
		while (re_node != null && re_node.llink != null)
			re_node = re_node.llink;
		return re_node;
	}

	// 搜尋左子樹替代節點
	public Student search_re_l(Student node)
	{
		Student re_node;
		re_node = node;
		while (re_node != null && re_node.rlink != null)
			re_node = re_node.rlink;
		return re_node;
	}

	// 搜尋node的父節點
	Student search_p(Student node)
	{
		Student parent;
		parent = root;
		while(parent != null) {
			if (node.name.compareTo(parent.name) < 0)
				if (node.name.compareTo(parent.llink.name) == 0)
					return parent;
				else
					parent = parent.llink;
			else
				if (node.name.compareTo(parent.rlink.name) == 0)
					return parent;
				else
					parent = parent.rlink;
		}
		return null;
	}

	public static void main(String args[])
	{
		int option = 0;
		Scanner keyboard = new Scanner(System.in);
		BinarySearchTree obj = new BinarySearchTree();

		while (true) {
			System.out.println("");
			System.out.println("*************************");
			System.out.println("        <1> insert          ");
			System.out.println("        <2> delete          ");
			System.out.println("        <3> modify          ");
			System.out.println("        <4> show            ");
			System.out.println("        <5> quit            ");
			System.out.println("*************************");
			System.out.print("Enter your choice: ");
 	 	    try {
				option = keyboard.nextInt();
			} catch(InputMismatchException e) {
				keyboard.nextLine();
				System.out.printf("Not a correctly number.\n");
				System.out.printf("Try again\n\n");
			}
			switch (option) {
				case 1:
					obj.insert_f();
					break;
				case 2:
					obj.delete_f();
					break;
				case 3:
					obj.modify_f();
					break;
				case 4:
					obj.show_f();
					break;
				case 5: 
					System.exit(0);
				default:
					System.out.println("Wrong option!");
			}
		}
	}
}
