package heap;

/**
 *
 * @author Bright
 * Version 2
 * Update date: March 20, 2017
 */

import java.io.*;
import java.util.Scanner;
import java.util.InputMismatchException;

class Heap {
	final int MAX=100;  // 定義此 Heap 之最大元素個數
	int[] heap_tree = new int[MAX]; // Heap陣列
	int last_index;     // 最後一筆資料的INDEX
	Scanner keyboard = new Scanner(System.in);

	Heap() // 建構函數
	{
		last_index=0;
	}
 
   // 新增函數
	public void insert_f() 
	{
	 	int id_temp=0;

		if (last_index >= MAX-1) { // 資料數超過上限，顯示錯誤訊息
			System.out.println("    \n   Login members are more than "+(MAX-1)+" !!");
			System.out.println("      Please wait for a Minute ...\n");
		}
		else {
			System.out.print("\n     Please enter login ID number: ");
			id_temp=keyboard.nextInt();
			 
			create(id_temp); // 建立Heap
			System.out.println("          Login successfully!!\n");
		}
	}

   // 刪除函數 
	public void delete_f() 
	{ 
		int id_temp=0,del_index=0;

		if (last_index<1) { // 無資料存在，顯示錯誤訊息
			System.out.println("\n        No member to logout!! ");
			System.out.println("        Please check again!! \n");
		}
		else {
			System.out.print("\n    Please enter logout ID number: ");
			id_temp=keyboard.nextInt();
             
			del_index = search(id_temp);
			if (del_index==0) { // 沒找到資料，顯示錯誤訊息
				System.out.println("         ID number not found!!\n");
 			}   
			else {
				removes(del_index);  // 刪除資料，並調整Heap
				System.out.println("  ID number " + id_temp + " logout!!\n");	 
			}
		}
	}
   
	// 輸出函數
	public void display_f() 
	{ 
	 	int option=0;

		if (last_index<1) // 無資料存在，顯示錯誤訊息
			System.out.println("\n        No member to show!!\n\n");
		else {
			System.out.println();
			System.out.println("*************************");
			System.out.println("*    <1> increase    *");// 選擇第一項為由小到大排列
			System.out.println("*    <2> decrease    *");// 選擇第二項為由大到小排列
			System.out.println("**************************");
			do {
				System.out.print("\n   Please enter your option: ");
 	 	        try {
				   option = keyboard.nextInt();
			    } catch(InputMismatchException e) {
				   keyboard.nextLine();
				   System.out.printf("Not a correctly number.\n");
				   System.out.printf("Try again\n\n");
			    }
			} while(option!= 1 && option != 2);	
			show(option);
		}
	}
	
   // 建立資料於Heap, ID_TEMP為新增資料
	public void create(int id_temp) 
	{ 
		heap_tree[++last_index]=id_temp; // 將資料新增於最後
		adjust_u(heap_tree, last_index); // 調整新增資料
	} 

	// 從Heap中刪除資料,INDEX_TEMP為欲刪除資料之INDEX
	public void removes(int index_temp) 
	{  
		// 以最後一筆資料代替刪除資料	 
		heap_tree[index_temp]=heap_tree[last_index];
		heap_tree[last_index--]=0;
		if (last_index>1) { // 當資料筆數大於1筆，則做調整
			// 當替代資料大於其PARENT NODE，則往上調整   
			if (heap_tree[index_temp]>heap_tree[index_temp/2] && index_temp>1)  
				adjust_u(heap_tree, index_temp);
			else // 替代資料小於其CHILDREN NODE，則往下調整
				adjust_d(heap_tree, index_temp, last_index-1);
		}     
	}
	
   // 印出資料於螢幕
	public void show(int op) 
	{
		int[] heap_temp=new int[MAX];
		int c_index=0;
		char tChar;

		// 將Heap資料複製到另一個陣列作排序工作
		for (c_index=1;c_index<=last_index;c_index++)
			heap_temp[c_index]=heap_tree[c_index];
		// 將陣列調整為由小到大排列
		for (c_index=last_index-1;c_index>0;c_index--) {
			exchange(heap_temp,1,c_index+1);		  
			adjust_d(heap_temp,1,c_index);
		}
		System.out.println("\n\n       ID number");
		System.out.println(" *********************");
		// 選擇第一種方式輸出，以遞增方式輸出--使用堆疊
		// 選擇第二種方式輸出，以遞減方式輸出--使用佇列
		switch(op) {
			case 1 :
				for (c_index = 1; c_index <= last_index; c_index++) 
					System.out.println(" "+heap_temp[c_index]); 	 
				break;
			case 2 :
				for (c_index = last_index; c_index > 0; c_index--)
					System.out.println(" "+heap_temp[c_index]);
				break;
		}
		System.out.println(" *********************");
		System.out.println(" Total member: "+last_index+"\n");
	} 
   
	// 從下而上調整資料, INDEX為目前資料在陣列之INDEX
	public void adjust_u(int temp[],int index) 
	{
		while(index>1) { // 將資料往上調整至根為止
			if (temp[index]<=temp[index/2])  // 資料調整完畢就跳出，否則交換資料
				break;
			else 			
				exchange(temp,index,index/2);
			index /= 2;
		}
	}

	// 從上而下調整資料, INDEX1為目前資料在陣列之INDEX，INDEX2為最後一筆資料在陣列之INDEX
	public void adjust_d(int temp[], int index1, int index2) 
	{  
		// ID_TEMP記錄目前資料，INDEX_TEMP則是目前資料之CHILDREN NODE的INDEX
		int id_temp=0, index_temp=0;

		id_temp = temp[index1];
		index_temp = index1 * 2;
		// 當比較資料之INDEX不大於最後一筆資料之INDEX，則繼續比較
		while (index_temp <= index2) {
			if ((index_temp < index2) && (temp[index_temp] < temp[index_temp+1]))
				index_temp++;  // INDEX_TEMP記錄目前資料之CHILDREN NODE中較大者
			if (id_temp >= temp[index_temp])  // 比較完畢則跳出，否則交換資料
				break;
			else {
				temp[index_temp/2] = temp[index_temp];
				index_temp *= 2;
			}
		}
		temp[index_temp/2] = id_temp;
	}

   // 交換傳來之ID1及ID2儲存之資料
	public void exchange(int arr[], int id1, int id2) 
	{ 
		int id_temp;

		id_temp = arr[id1];
		arr[id1]=arr[id2];
		arr[id2]=id_temp;	  
	}
	
   //  尋找陣列中ID_TEMP所在
	public int search(int id_temp) 
	{ 
		int c_index=0;

		for (c_index=1;c_index<=MAX-1;c_index++) 
			if(id_temp==heap_tree[c_index])
				return c_index; // 找到則回傳資料在陣列中之INDEX
		return 0; // 沒找到則回傳0
	}

	public static void main (String args[]) 
	{ 	 
		Scanner keyboard = new Scanner(System.in);
 		int option=0;
		Heap obj = new Heap();

		do {			 
			System.out.println("******** HeapTree Program  ********");
			System.out.println("          <1> Login                ");
			System.out.println("          <2> Logout               ");
			System.out.println("          <3> Show                 ");
			System.out.println("          <4> Exit                 ");
 			System.out.println("***********************************");
			System.out.print("\n             Choice :              "); 
 	 	    try {
				option = keyboard.nextInt();
			} catch(InputMismatchException e) {
				keyboard.nextLine();
				System.out.printf("Not a correctly number.\n");
				System.out.printf("Try again\n\n");
			}
			switch(option) {			   
				case 1 :
					obj.insert_f();  // 新增函數
					break;
				case 2 :
					obj.delete_f();  // 刪除函數
					break;
				case 3 :
					obj.display_f(); // 輸出函數
					break;			      
				case 4 :
					System.exit(0);	
			}
		} while (true);	 
	}
}
