package avltree;

/**
 *
 * @author Bright
 * Version 2
 * Update date: March 22, 2017
 */


import java.io.*;
import java.util.Scanner;
import java.util.InputMismatchException;

// 定義一個 Node 的資料結構(Class)，其資料含兩個子鏈結，儲存姓名、分數、節點BF值              
class student { 
	public String name; // 姓名
	public int score; // 分數
	public int bf; // 節點BF值
	public student llink,rlink ; //節點子鏈結
}

class AVLTree {
	student ptr, root, current, prev, pivot, pivot_prev;
	Scanner keyboard = new Scanner(System.in);
	int nodecount = 0; /* 用來計算node個數 */

	AVLTree() { // 建構子
		root=prev=pivot_prev=null;
	}

	// 新增函數
	public void insert_f() 
	{ 
	 	String name_t="";
		int score_t=0;

		System.out.printf("\n************ Insert Node ************\n");
		System.out.print("   Please enter student name  : ");
  		name_t = keyboard.next();
	     			
		System.out.print("   Please enter student score : ");
 		score_t = keyboard.nextInt();
           
		System.out.println("\n               New Node            ");
		System.out.println("-------------------------------------");
		System.out.println("       Name : " + name_t + "     Score : " + score_t);
		System.out.println("*************************************\n");
		nodecount++;/* 將node加1 */
		sort_f(name_t,score_t);   
	}
	
	// 處理 AVL-TREE 之資料輸入
	public void sort_f(String name_t,int score_t) 
	{ 	
		int op;
		current = root;
		while ((current != null) && !name_t.equals(current.name)) {
			if (name_t.compareTo(current.name)<0) { // 插入資料小於目前位置，則往左移 
				prev = current;
				current = current.llink;
			}
			else { // 若大於目前位置，則往右移 
				prev = current;
				current = current.rlink;
			}
		}
		// 找到插入位置，無重覆資料存在
		if ((current==null) || !name_t.equals(current.name)) { 
			ptr = new student();  // 配置記憶體
			ptr.name = name_t;
			ptr.score = score_t;
			ptr.llink = null;
			ptr.rlink = null;
			if (root == null) // ROOT不存在，則將ROOT指向插入資料 
				root = ptr;   
			else
				if (ptr.name.compareTo(prev.name)<0) 
					prev.llink = ptr;
				else
					prev.rlink = ptr;
			bf_count(root);
			pivot = pivot_find();
			if (pivot != null) { // PIVOT存在，則須改善為AVL-TREE 
				op = type_find();
				switch(op) {
					case 11 :
						type_ll();
						break;
					case 22 :
						type_rr();
						break;
					case 12 :
						type_lr();
						break;
					case 21 :
						type_rl();
						break;
				}
			}
			bf_count(root);  // 重新計算每個節點的BF值 
		}
		else {  // 欲插入資料KEY已存在，則顯示錯誤
			System.out.println("      ADD  New Node  error !!");
			System.out.println("      student <" + name_t + "> has existed !");
		}
	}

   // 計算BF值，使用後序法逐一計算 
	public void bf_count(student trees) 
	{
		if (trees != null) {
			bf_count(trees.llink);
			bf_count(trees.rlink);
			// BF值計算方式為左子樹高減去右子樹高 
			trees.bf = height_count(trees.llink) - height_count(trees.rlink);
		}
	}

	public int height_count(student trees) 
	{
		if (trees == null) 
			return 0;
		else
			if (trees.llink == null && trees.rlink == null)
				return 1;
			else 
				return 1 + (height_count(trees.llink) > height_count(trees.rlink) ?
						height_count(trees.llink) : height_count(trees.rlink));
	}

	public student pivot_find() 
	{
		int i;
		current = root;
		pivot = null;
		for (i =0; i<=nodecount-1; i++) {
			// 當BF值的絕對值小於等於1，則將PIVOT指向此節點 
			if (current.bf < -1 || current.bf > 1) {
				pivot = current;
				if(pivot != root) 
					pivot_prev = prev;
				System.out.println("current pivot name: " + current.name);
			}
			if (current.bf > 0) { /* 左子樹的高度較高 */
				prev = current;
				current = current.llink;
			}
			else if (current.bf < 0 ) { /* 右子樹的高度較高 */
				prev = current;
				current = current.rlink;
			}
		}
		return pivot;
	}

	public int type_find() 
	{
		int i, op_r = 0;
		current = pivot;
		for (i = 0; i < 2; i++) {
			if (current.bf > 0) { /* 左子樹的高度較高 */
				current = current.llink;
				if (op_r == 0) 
					op_r+=10;
				else 
					op_r++;
			}
			else if (current.bf < 0 ) { /* 右子樹的高度較高 */
				current = current.rlink;
				if (op_r == 0)
					op_r+=20;
				else
					op_r+=2;
			}
		}
		// 傳回值11、22、12、21分別代表LL、RR、LR、RL型態  
		return op_r;
	}
	
   // LL型態
	public void type_ll() 
	{  
		student pivot_next, temp;

		pivot_next = pivot.llink;
		temp = pivot_next.rlink;
		pivot_next.rlink = pivot;
		pivot.llink = temp;
		if (pivot == root)
			root = pivot_next;
		else
			if (pivot_prev.llink == pivot)
				pivot_prev.llink = pivot_next;
			else
				pivot_prev.rlink = pivot_next;
	}
	
   // RR型態 
	public void type_rr() 
	{ 
		student pivot_next, temp;

		pivot_next = pivot.rlink;
		temp = pivot_next.llink;
		pivot_next.llink = pivot;
		pivot.rlink = temp;
		if (pivot == root) 
			root = pivot_next;
		else
			if (pivot_prev.llink == pivot)
				pivot_prev.llink = pivot_next;
			else
				pivot_prev.rlink = pivot_next;
	}

   // LR型態
	public void type_lr() 
	{  
		student pivot_next, temp;

		pivot_next = pivot.llink;
		temp = pivot_next.rlink;
		pivot.llink = temp.rlink;
		pivot_next.rlink = temp.llink;
		temp.llink = pivot_next;
		temp.rlink = pivot;
		if (pivot == root)
			root = temp;
		else
			if (pivot_prev.llink == pivot)
				pivot_prev.llink = temp;
			else
				pivot_prev.rlink = temp;
	}

   // RL型態
	public void type_rl() 
	{ 
		student pivot_next, temp;

		pivot_next = pivot.rlink;
		temp = pivot_next.llink;
		pivot.rlink = temp.llink;
		pivot_next.llink = temp.rlink;
		temp.rlink = pivot_next;
		temp.llink = pivot;
		if (pivot == root)
			root = temp;
		else 
			if (pivot_prev.llink == pivot)
				pivot_prev.llink = temp;
			else
				pivot_prev.rlink = temp;
	}
	
   //輸出函數
	public void list_f() 
	{ 
		 
		if (root==null) { //無資料存在，則顯示錯誤
			System.out.println("\n      No Student record exist !!");
		}
		else {
			list_data(); //中序列印
		}
 	}		 
	 
   //將 Node 資料以中序方式印出
	public void list_data() 
	{ 
		System.out.println("\n\n******** List Data *********\n");
		System.out.printf("%-15s %-10s\n", "Name", "Score");
		System.out.println("-----------------------------");
		inorder(root);
		System.out.println("-----------------------------");
	}	
	
	// 中序使用遞迴
	public void inorder(student trees) 
	{  
		try {
			inorder(trees.llink);			
 			System.out.printf("%-16s", trees.name);			
 			System.out.printf("%3d\n", trees.score);
			inorder(trees.rlink);
	   } catch (NullPointerException e) {}
	}

	//修改函數
	public void modify_f() 
	{ 
 		String name_t="";
		int score_t=0;

		if (root==null) { //無資料存在，則顯示錯誤
			System.out.println("\n      No Student record exist !!");
		}
		else {
			System.out.printf("\n************ Modify Node ************\n");
			System.out.print("   Please enter student name  : ");	
 			name_t = keyboard.next();
 
			current=root;
			while ((current != null) && (!name_t.equals(current.name))) {
				if (name_t.compareTo(current.name) < 0)
					current = current.llink;
				else
					current = current.rlink;
			}
			// 若找到欲更改資料，則列出原資料，並要求輸入新的資料  
			if (current != null) {
				System.out.println("\n          Student name  : " +current.name);
				System.out.println("          Studnet score : " +current.score);
				System.out.println("\n*************************************");
				System.out.print("   Please enter new score : ");
		 		current.score = keyboard.nextInt();
 
				System.out.printf("    Data updated successfully");
			}				 
			else { // 找不到資料，則顯示錯誤
				System.out.println("\n    Student : " +name_t+ "  Not Found !!!");
			}
		}
		System.out.println("\n");
	}	

   //刪除函數
	public void delete_f() 
	{ 
 		student clear; 
		int op;
		String name_t="", tempn="";	   
		 
		if (root == null) { //無資料存在，則顯示錯誤
			System.out.println("\n      No Student record exist !!");
 		}
		else {
			System.out.println("\n************ Delete Node ************\n");
			System.out.print("   Please enter student name  : ");		
			name_t = keyboard.next();
			tempn = name_t;
			current = root;
			//尋找刪除點
			while (current != null && !name_t.equals(current.name)) {
				// 若刪除資料鍵值小於目前所在資料，則往左子樹 
				if (name_t.compareTo(current.name)<0) {
					prev = current;
					current = current.llink;
				}
				// 否則往右子樹 
				else {
					prev = current;				  			
					current = current.rlink;				  
				}
			}
			// 找到欲刪除資料的狀況 			 
			if (current!=null && name_t.equals(current.name)) {
				// 當欲刪除資料底下無左右子樹存在的狀況 
				if (current.llink == null && current.rlink == null) {
					clear = current;
					if (name_t.equals(root.name)) // 欲刪除資料為ROOT 
						root = null;
					else {
						// 若不為ROOT，則判斷其為左子樹或右子樹 
						if (name_t.compareTo(prev.name) < 0)
							prev.llink = null;
						else
							prev.rlink = null;
					}    
					clear = null;              
				}				
				else {
					// 以左子樹最大點代替刪除資料 				
					if (current.llink != null) {				
						clear = current.llink;
						while (clear.rlink != null) {
							prev = clear;
							clear = clear.rlink;
						}
						current.name=clear.name;
						current.score = clear.score;
						if (current.llink == clear)
							current.llink = clear.llink;
						else
							prev.rlink = clear.llink;
					}
					else {  // 以右子樹最小點代替刪除資料 
						clear = current.rlink;
						while (clear.llink != null) {
							prev = clear;
							clear = clear.llink;
						}						   
						current.name=clear.name;
						current.score = clear.score;
						if (current.rlink == clear)
							current.rlink = clear.rlink;
						else
							prev.llink = clear.rlink;
					}
					clear = null;				
				}				
				bf_count(root);
				if (root != null) { // 若ROOT不存在，則無需作平衡改善       
					pivot = pivot_find();      // 尋找PIVOT所在節點 
					while (pivot != null) {
						op = type_find();
						switch(op) {
							case 11 : 
								type_ll();
								break;
							case 22 :
								type_rr();
								break;
							case 12 :
								type_lr();
								break;
							case 21 :
								type_rl();
								break;
						}
						bf_count(root);
						pivot = pivot_find();
					}
				}
				nodecount--; /* 將node減1 */
				System.out.printf("\n   Student %s has been deleted\n", tempn);
			}        
			else { //找不到資料，則顯示錯誤
				System.out.printf("\n    Student %s not found !!!\n", tempn);
			}
		}
	}

   //主函數
	public static void main (String args[]) 
	{ 		 
		Scanner keyboard = new Scanner(System.in);
		int option=0;  
	   
		AVLTree obj = new AVLTree();
		do {	
	       System.out.printf("\n");		 
			System.out.println("*******  AVLtree Program *********");
			System.out.println("                                  ");
			System.out.println("         <1> Insert Node          ");
			System.out.println("         <2> Delete Node          ");
			System.out.println("         <3> Modify Node          ");	 
			System.out.println("         <4> List   Node          ");
			System.out.println("         <5> Exit                 ");
			System.out.println("                                  ");	 
			System.out.println("**********************************");
			System.out.print("\n             Choice :             ");		 
			try {
				option = keyboard.nextInt();
			} catch(InputMismatchException e) {
				keyboard.nextLine();
				System.out.printf("Not a correctly number.\n");
				System.out.printf("Try again\n\n");
			}
			switch(option) {			   
				case 1 :
					obj.insert_f();  //新增函數
					break;
				case 2 :
					obj.delete_f();  //刪除函數
					break;
				case 3 :
					obj.modify_f();   //修改函數
					break;
				case 4 :
					obj.list_f();    //輸出函數
					break;
				case 5 :
					System.exit(0);	
			}			
		} while (true);
	}
}
