package mwaysearchtree;

/**
 *
 * @author Bright
 * Version 2
 * Update date: March 22, 2017
 */

import java.io.*;
import java.util.Scanner;
import java.util.InputMismatchException;

class TreeNode {       // 定義資料欄位
	public char id;     // 輸出時識別節點
	public int n;       // 節點內的鍵值數
	public int[] key=new int[3];    // 節點鍵值
	public TreeNode[] link=new TreeNode[3];    // 節點子鏈結
} 

class MWaySearchTree {
	final int MAX=3; // 設定此為3-WAY TREE的最大節點數
	TreeNode ptr,root,node,prev,parent,replace;
	char id_seq='\0';
	Scanner keyboard = new Scanner(System.in);
       
	MWaySearchTree() {  // 建構子
		root=null;
	}
 
   // 加入函數--新增一筆資料
	public void insert_f() 
	{
    	int add_num=0;

		System.out.print("\n    Please enter insert number: ");
 		add_num = keyboard.nextInt();
 		create(add_num);
 		System.out.printf("\n");
	}

   // 刪除函數--刪除一筆資料
	public void delete_f() 
	{ 
 		int del_num=0, ans=0;
   
		if (root == null)  // 當樹根為NULL，顯示錯誤訊息
			System.out.println("\n           No data found!!\n");
		else {
			System.out.printf("\n Please enter delete number: ");
 			del_num = keyboard.nextInt();
 			
			ans = search_num(del_num);  // 搜尋資料是否存在
			if (ans == 0)  // 當資料不存在，顯示錯誤訊息
				System.out.println("Number " + del_num + " not found!!\n");
			else {
				removes(node, ans);
				System.out.println("Number " + del_num + " deleted!!\n");
			}
		}
	}
     
   // 輸出函數--將M-WAY TREE內的所有資料輸出     
	public void display_f() 
	{ 
		if (root == null)  // 當樹根為NULL，顯示錯誤訊息
			System.out.println("\n           No data found!!\n");
		else {
			id_seq = 'a';  // 節點編號由a開始
			preorder_id(root);  // 給予每個節點編點
			System.out.printf(" \n The data of M-way search tree is listing below: ");
			System.out.print("\n =======================================\n");
			preorder_num(root); // 輸出節點資料
			System.out.println(" =======================================\n");
		}
	}

	// 將資料加入，並調整為M-WAY TREE，NUM為新增之資料鍵值
	public void create(int num) 
	{
		int ans = 0, i = 0;

		if (root == null) { // 樹根為NULL的狀況
			initial();
			ptr.key[1] = num;
			ptr.n++;
			root = ptr;
		}
		else {
			ans = search_num(num);  // 搜尋資料是否已存在
			if (ans != 0)  // 資料存在，則顯示錯誤訊息
				System.out.println("    Number " + num + " has existed!!\n");
			else {
				node = search_node(num);  // 找尋插入點
				if (node != null) { // 插入點還有空間存放資料的狀況
					for (i = 1; i < MAX-1; i++)
						if (num < node.key[i])
							break;
					moveright(i, num);
				}
				else { // 新增加一個節點加入資料的狀況
					initial();
					ptr.key[1] = num;
					ptr.n++;
					for (i = 1; i < MAX; i++)
						if (num < prev.key[i])
							break;
					prev.link[i-1] = ptr;
				}
			} 
		}
		System.out.printf("%10d has been inserted!\n", num);
	}

	// 將資料移除，並調整為M-WAY TREE，node_TEMP為刪除資料所在節點，
	// LOCATION為資料在節點中的位置
	public void removes(TreeNode node_temp, int location) 
	{
		int i;

		node = node_temp;
		replace = find_next(node.link[location]); // 找尋替代之後繼節點
		if (replace == null) { // 沒有後繼節點的狀況
			replace = find_prev(node.link[location-1]); // 找尋替代之前行節點
			if (replace == null) { // 沒有前行節點的狀況    
				moveleft(location);
				replace = node;
				if (node.n == 0) { // 刪除資料後，節點內資料為0的處理
					if (node == root)  // 當節點為根的狀況
						root = null;
					else  // 節點不是根，則調整鏈結
						for (i = 0; i <= parent.n; i++)
							if (node == parent.link[i]) {
								parent.link[i] = null;
								break;
							}
				}
			}
			else {    // 有前行節點的狀況    
				// 以前行節點的資料代替刪除資料
				node.key[location] = replace.key[replace.n];
				parent = prev;
				removes(replace, replace.n);  // 移除替代資料
			}
		}
		else {    // 有後繼節點的狀況
			// 以後繼節點的資料代替刪除資料
			node.key[location] = replace.key[1];
			parent = prev;
			removes(replace, 1);  // 移除替代資料
		}
	}

	// 初始化節點--新增一個節點，將其所有鏈結指向NULL，設其節點數為0
	public void initial() 
	{
		int i=0;

		ptr = new TreeNode();
		for (i = 0; i < MAX; i++)
			ptr.link[i] = null;
		ptr.n = 0;
	}

	// 搜尋節點位置--搜尋NUM，存在則回傳NUM在節點中的位置，不存在則回傳0
	public int search_num(int num) 
	{
		int n_temp=0, done=0, i=0;
 
		node = root;
		while (node != null) {
			parent = prev;
			prev = node;
			for (i = 1, done = 0; i <= node.n; i++) {
				if (num == node.key[i])
					return i;   // 找到NUM，回傳其在節點中的位置
				if (num < node.key[i]) {
					node = node.link[i-1];
					done = 1;
					break;
				}
			}
			if (done == 0)
				node = node.link[i-1];
		}
		return 0;   // 沒有找到則回傳0
	}

	// 搜尋節點--尋找插入NUM的節點，並回傳插入節點
	public TreeNode search_node(int num) 
	{
		int i=0, done=0;
		TreeNode node_temp;

		node_temp = root;
		while(node_temp != null) {
			if (node_temp.n < MAX-1)
				return node_temp;  // 找到有多餘空間存放NUM，則回傳此節點
			else {
				for (i = 1, done = 0; i < MAX; i++) {
					if (node_temp.n < i)
						break;
					if (num < node_temp.key[i]) {
						node_temp = node_temp.link[i-1];
						done = 1;
						break;
					}
				}
				if (done == 0)
					node_temp = node_temp.link[i-1];
			}
		}
		return node_temp;  // 若沒有找到有多餘空間存放NUM的節點，回傳NULL
	}

	// 將節點內資料右移--將節點資料右移至INDEX位置，並將NUM插入
	public void moveright(int index, int num) 
	{
		int i=0;

		for (i = node.n+1; i > index; i--) { // 資料右移至INDEX處
			node.key[i] = node.key[i-1];
			node.link[i] = node.link[i-1];
		}
		node.key[i] = num;  // 插入NUM
		// 調整NUM左右鏈結
		if (node.link[i-1] != null && node.link[i-1].key[0] > num) {
			node.link[i] = node.link[i-1];
			node.link[i-1] = null;
		}
		node.n++;
	}

	// 將節點內資料左移--將節點資料從INDEX位置左移
	public void moveleft(int index) 
	{
		int i=0; 

		for (i = index; i < node.n; i++) { // 節點資料左移
			node.key[i] = node.key[i+1];
			node.link[i] = node.link[i+1];
		}
		node.n--;
	}

	// 尋找後繼節點--尋找node_TEMP的後繼節點，回傳找到的後繼節點
	public TreeNode find_next(TreeNode node_temp) 
	{
		prev = node;
		if (node_temp != null)
			while (node_temp.link[0] != null) {
				prev = node_temp;
				node_temp = node_temp.link[0];
			}
		return node_temp;
	}

	// 尋找前行節點--尋找node_TEMP的前行節點，回傳找到的前行節點
	public TreeNode find_prev(TreeNode node_temp) 
	{
		prev = node;
		if (node_temp != null)
			while (node_temp.link[MAX-1] != null) {
				prev = node_temp;
				node_temp = node_temp.link[MAX-1];
			}
		return node_temp;
	}

	// 給予節點編號--使用前序遞迴方式給予每個節點編號
	public void preorder_id(TreeNode tree) 
	{
		int i=0;

		if (tree != null) {
			tree.id = id_seq++;
			for (i = 0; i <= tree.n; i++)
				preorder_id(tree.link[i]);
		}
	}
 
	// 輸出資料--使用前序遞迴方式輸出節點資料
	public void preorder_num(TreeNode tree) 
	{
		int i=0;
		char link_id='\0';

		if (tree != null) {
			// 當節點鏈結為NULL，則顯示鏈結為0
			if (tree.link[0] == null) 
				link_id = '0';
			else
				link_id = tree.link[0].id;
			System.out.print("  " + tree.id + ", " + tree.n + ", " + link_id);
			for (i = 1; i <= tree.n; i++) {
				if (tree.link[i] == null) 
					link_id = '0';
				else
					link_id = tree.link[i].id;
				System.out.print(", (" + tree.key[i] + ", " + link_id + ")");
			}
			System.out.println("");
			for (i = 0; i <= tree.n; i++)
				preorder_num(tree.link[i]);
		}
	}

   // 主函數
	public static void main (String args[]) 
	{ 
		Scanner keyboard = new Scanner(System.in);
 		int option=0;
		  
		MWaySearchTree obj = new MWaySearchTree();
		do {			 
			System.out.println("****** m-way search tree ******");
			System.out.println("                               ");
			System.out.println("          <1> Login            ");
			System.out.println("          <2> Logout           ");
			System.out.println("          <3> Show             ");
			System.out.println("          <4> Exit             ");
			System.out.println("*******************************");
			System.out.print("\n           Choice :            ");	
			try {
				option = keyboard.nextInt();
			} catch(InputMismatchException e) {
				keyboard.nextLine();
				System.out.printf("Not a correctly number.\n");
				System.out.printf("Try again\n\n");
			}
			switch(option) {
				case 1 :
					obj.insert_f();  // 新增函數
					break;
				case 2 :
					obj.delete_f();  // 刪除函數
					break;
				case 3 :
					obj.display_f(); // 輸出函數
					break;
				case 4 :
					System.exit(0);
			}
		} while (true);
	}
}

