package dijkstra;

/**
 *
 * @author Bright
 * Version 2
 * Update date: March 23, 2017
 */

import java.io.*;
import java.util.*;

class Dijkstra {
	final int MAX_V = 100;
	final int VISITED = 1;
	final int NOTVISITED = 0;
	final int Infinite = 1073741823;
	// A[1..N][1..N] 為圖形的相鄰矩陣
	// D[i] i=1..N 用來儲存某起始頂點到i節點的最短距離
	// S[1..N] 用來記錄頂點是否已經拜訪過
	// P[1..N] 用來記錄最近經過的中間節點
	int[][] A = new int[MAX_V+1][MAX_V+1];
	int[] D = new int[MAX_V+1];
	int[] S = new int[MAX_V+1];
	int[] P = new int[MAX_V+1];
	int source, sink, N;
	int step;
	int top;         // 堆疊指標
	int[] Stack = new int[MAX_V+1];   // 堆疊空間
    Scanner inputStream = null;
    Scanner keyboard = new Scanner(System.in);
 
	public Dijkstra()
	{
		step = 1;
		top = -1;
	}

	public void init()
	{
		int i, j = 0, weight = 0;
		boolean done = false;
		
		try {
			inputStream = new Scanner(new FileInputStream("sh_path.dat"));
		} catch(FileNotFoundException e) {
			System.err.println("File sh_path.dat not found!");
			System.exit(1);
		}

		// 讀取圖形節點數
		try {
			N = inputStream.nextInt();
		} catch(NoSuchElementException e) {}

		for ( i=1; i<=N; i++ )
			for ( j=1; j<=N; j++ )
				A[i][j] = Infinite;  // 起始A[1..N][1..N]相鄰矩陣

		while ( done == false ) {
			try {
				i = inputStream.nextInt();
				j = inputStream.nextInt();
				weight = inputStream.nextInt();
			} 
			catch(NoSuchElementException e) {	
				done = true;
			}
			A[i][j] = weight;    // 讀取i節點到j節點的weight
		}
 
		inputStream.close();
 
		System.out.printf("Enter source node : ");
		source = keyboard.nextInt();
		System.out.printf("Enter sink node : ");
		sink = keyboard.nextInt();

		// 起始各陣列初始值
		for ( i = 1; i <= N; i++ ) {
		   /* 各頂點設為尚未拜訪記錄起始頂點至各頂點最短距離 */
		   S[i] = NOTVISITED;   
		   D[i] = A[source][i];    
		   P[i] = source;
		}
		// 始起節點設為已經走訪
		S[source] = VISITED;    
		D[source] = 0;
	}

	public void access()
	{
		int I, t;
		for ( step =2;step <=N; step++ ) {
		   // minD 傳回一值t使得D[t] 為最小

			t = minD();
			S[t] = VISITED;

			// 找出經過t點會使路徑縮短的節點
			for ( I=1; I <= N; I++ )
				if ( (S[I] == NOTVISITED) && (D[t]+A[t][I] <= D[I]) ) {
					D[I] = D[t] + A[t][I];
					P[I] = t;
				}
			output_step();
		}
	}

	public int minD()
	{
		int i,t = 0;
		int minimum = Infinite;
		for ( i=1;i<=N;i++ )
			if ( (S[i] == NOTVISITED) && D[i] < minimum )
			{
				minimum = D[i];
				t = i;
			}
		return t;
	}

	// 顯示目前的D陣列與P陣列狀況
	public void output_step()
	{
		int i;

		System.out.print("\n Step #" + step);
		System.out.println("\n================================================");
		for ( i=1; i<=N; i++ )
			System.out.printf(" D[%d] ", i);
		System.out.println("");

		for ( i=1; i<=N; i++ )
			if ( D[i] == Infinite )
				System.out.print(" ---- ");
			else
				System.out.printf(" %3d  ", D[i]);

		System.out.println("\n================================================");
		for ( i=1; i<=N; i++ )
			System.out.printf(" P[%d] ", i);
		System.out.println("");

		for ( i=1; i<=N; i++ )
			System.out.printf(" %3d  ", P[i]);
	}

	// 顯示最短路徑
	public void output_path()
	{
		int node = sink;

		// 判斷是否起始頂點等於終點或無路徑至終點
		if ( (sink == source) || (D[sink] == Infinite) ) {
			System.out.print("\nNode " + source +
								" has no Path to Node " + sink);
			return;
		}

		System.out.printf("\n\n");
		System.out.print(" The shortest  Path from V" + source +
							" to V" + sink + " :");
		System.out.println("\n------------------------------------------");

		// 由終點開始將上一次經過的中間節點推入堆疊直到起始節點
		System.out.print("  V" + source);
		while ( node != source ) {
		   Push(node);
		   node  = P[node];
		}
		while( node != sink) {
		   node = Pop();
		   System.out.print(" --" + A[P[node]][node] + "-->");
		   System.out.print("V" + node);
		}

		System.out.println("\n Total length : " + D[sink]);
	}

	public void Push(int value)
	{
		if ( top >= MAX_V ) {
			System.out.println("Stack overflow!");
			System.exit(1);
		}
		else
			Stack[++top] = value;
	}

	public int Pop()
	{
		if ( top < 0 ) {
			System.out.println("Stack empty!");
			System.exit(1);
		}
			return Stack[top--];
	}

	public static void main(String args[])
	{
		Dijkstra obj = new Dijkstra();
		
 		obj.init();
		obj.output_step();
		obj.access();
		obj.output_path();
	}
}
